/**
 * NASR - Noise-Adaptive Shape Reconstruction
 * Source code, Version 1.0 (December 2014)
 *
 * Copyright (c) 2014 Simon Giraudot <simon.giraudot@inria.fr>
 * http://www-sop.inria.fr/members/Simon.Giraudot/
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#ifndef __DT__
#define __DT__

#include <queue>
#include <list>
#include <algorithm>
#include <iostream>
#include <fstream>

#include <omp.h>

#include <CGAL/Delaunay_triangulation_3.h>
#include <CGAL/iterator.h>
#include <CGAL/centroid.h>
#include <CGAL/convex_hull_3.h>
#include <CGAL/intersections.h>

#include "dpqueue.h"

template < typename Kernel,
	   typename Cb = CGAL::Triangulation_cell_base_3<Kernel> >
class My_cell_base : public Cb
{

public:

  typedef typename Kernel::FT FT;
  typedef typename Kernel::Point_3 Point;
  typedef typename Kernel::Vector_3 Vector;
  typedef typename Kernel::Plane_3 Plane;
  typedef typename CGAL::Tetrahedron_3<Kernel> Tetrahedron;
  typedef typename Cb::Cell_handle Cell_handle;
  typedef typename Cb::Vertex_handle Vertex_handle;


public:

  template < typename TDS2 >
  struct Rebind_TDS {
    typedef typename Cb::template Rebind_TDS<TDS2>::Other Cb2;
    typedef My_cell_base<Kernel,Cb2> Other;
  };
  
  My_cell_base() : Cb () { }
  My_cell_base(Cell_handle c) : Cb(c) { }
  My_cell_base(Vertex_handle v1,
	       Vertex_handle v2,
	       Vertex_handle v3,
	       Vertex_handle v4)
    : Cb(v1,v2,v3,v4) { }
  
  Point get_center ()
  {
    double x = 0, y = 0, z = 0;
    for (int i = 0; i < 4; ++ i)
      {
	x += this->vertex (i)->point ().x ();
	y += this->vertex (i)->point ().y ();
	z += this->vertex (i)->point ().z ();
      }
    return Point (x / 4, y / 4, z / 4);
  }

  Point get_random_point_inside_cell ()
  {
    Vector v01 (this->vertex (0)->point (), this->vertex (1)->point ());
    Vector v02 (this->vertex (0)->point (), this->vertex (2)->point ());
    Vector v03 (this->vertex (0)->point (), this->vertex (3)->point ());

    Point out = this->vertex (0)->point ();

    FT r01 = rand () / (FT)RAND_MAX;
    FT r02 = rand () / (FT)RAND_MAX;
    FT r03 = rand () / (FT)RAND_MAX;

    if (r01 + r02 > 1.)
      {
    	r01 = 1. - r01;
    	r02 = 1. - r02;
      }

    if (r01 + r02 + r03 > 1.)
      {
    	if (r02 + r03 > 1.)
    	  {
    	    FT tmp = r02;
    	    r02 = 1. - r03;
    	    r03 = 1. - r01 - tmp;
    	  }
    	else
    	  {
    	    FT tmp = r03;
    	    r03 = r01 + r02 + r03 - 1.;
    	    r01 = 1. - r02 - tmp;
    	  }
      }

    out = out + r01 * v01;
    out = out + r02 * v02;
    out = out + r03 * v03;

    return out;
  }


  bool contains (const Point& query)
  {
    Tetrahedron tetra (this->vertex (0)->point (),
		       this->vertex (1)->point (),
		       this->vertex (2)->point (),
		       this->vertex (3)->point ());
    return tetra.has_on_bounded_side (query);
  }
  


  FT volume ()
  {
    Point p[4];
    for (int i = 0; i < 4; ++ i)
      p[i] = this->vertex (i)->point ();

    Tetrahedron tetra (p[0], p[1], p[2], p[3]);
    return std::fabs (tetra.volume ());
  }



  FT distance_interpolated (const Point& query) const
  {
    const Vertex_handle v0 = this->vertex (0);
    const Vertex_handle v1 = this->vertex (1);
    const Vertex_handle v2 = this->vertex (2);
    const Vertex_handle v3 = this->vertex (3);
    const Point& p0 = v0->point ();
    const Point& p1 = v1->point ();
    const Point& p2 = v2->point ();
    const Point& p3 = v3->point ();

    return (std::fabs (CGAL::volume (p1, p2, p3, query) * v0->distance ()) +
	    std::fabs (CGAL::volume (p0, p2, p3, query) * v1->distance ()) +
	    std::fabs (CGAL::volume (p0, p1, p3, query) * v2->distance ()) +
	    std::fabs (CGAL::volume (p0, p1, p2, query) * v3->distance ()))
      / std::fabs (CGAL::volume (p0, p1, p2, p3));
  }



  FT implicit_interpolated (const Point& query) const
  {
    const Vertex_handle v0 = this->vertex (0);
    const Vertex_handle v1 = this->vertex (1);
    const Vertex_handle v2 = this->vertex (2);
    const Vertex_handle v3 = this->vertex (3);
    const Point& p0 = v0->point ();
    const Point& p1 = v1->point ();
    const Point& p2 = v2->point ();
    const Point& p3 = v3->point ();

    return (std::fabs (CGAL::volume (p1, p2, p3, query)) * v0->implicit () +
	    std::fabs (CGAL::volume (p0, p2, p3, query)) * v1->implicit () +
	    std::fabs (CGAL::volume (p0, p1, p3, query)) * v2->implicit () +
	    std::fabs (CGAL::volume (p0, p1, p2, query)) * v3->implicit ())
      / std::fabs (CGAL::volume (p0, p1, p2, p3));
  }



  FT distance_and_k (const Point& query, FT& distance, unsigned int& k)
  {
    Vertex_handle v[4];
    Point p[4];
    for (int i = 0; i < 4; ++ i)
      {
	v[i] = this->vertex (i);
	p[i] = v[i]->point ();
      }

    Tetrahedron tetra (p[0], p[1], p[2], p[3]);
    FT volume = tetra.volume ();
    distance = 0.;
    FT double_k = 0.;

    for (int i = 0; i < 4; ++ i)
      {
	Tetrahedron t (query, v[(i+1)%4]->point (), v[(i+2)%4]->point (),
		    v[(i+3)%4]->point ());
	distance += v[i]->distance () * std::fabs (t.volume () / volume);	
	double_k += v[i]->k () * std::fabs (t.volume () / volume);	
      }
    k = (int)double_k;
  }

};


template <class FT, class Cell_handle>
class CPCell
{
protected:

  Cell_handle m_cell;
  FT m_error;

public:


  CPCell ()
  {
    m_cell = Cell_handle ();
    m_error = 0.;
  }



  CPCell (const Cell_handle& ce, const FT error = 0.)
  {
    m_cell = ce;
    m_error = error;
  }



  CPCell (const CPCell& pcell)
  {
    m_cell = pcell.cell ();
    m_error = pcell.error ();
  }

  virtual ~CPCell () { }


  CPCell& operator= (const CPCell& pcell)
  {
    m_cell = pcell.cell ();
    m_error = pcell.error ();
    return *this;
  }


  bool operator== (const CPCell& pcell) const
  {
    return m_cell = pcell.cell ();
  }



  bool operator< (const CPCell& pcell) const
  {
    return m_cell < pcell.cell ();
  }

  const Cell_handle cell () const { return m_cell; }
  const FT error () const { return m_error; }

};


template <class T>
class CDPQueue : public DynamicPriorityQueue<T>
{
public:
  CDPQueue () { }
  ~CDPQueue () { }
  bool compare (const T& a, const T& b) const
  {
    return a.error () > b.error ();
  }
};



template < typename Kernel,
	   typename Vb = CGAL::Triangulation_vertex_base_3<Kernel> >
class My_vertex_base : public Vb
{
public:

  typedef typename Kernel::FT FT;
  typedef typename Vb::Point Point;
  typedef typename Kernel::Vector_3 Vector;
  typedef typename Vb::Cell_handle Cell_handle;



private:

  int m_index;
  int m_k;
  double m_distance;
  double m_implicit;
  double m_volume_around;
  bool m_constrained;
  double m_constraint;

public:

  template < typename TDS2 >
  struct Rebind_TDS {
    typedef typename Vb::template Rebind_TDS<TDS2>::Other Vb2;
    typedef My_vertex_base<Kernel, Vb2>   Other;
  };
  
  My_vertex_base() : Vb()
  {
    m_index = -1;
    m_k = -1;
    m_distance = 0.;
    m_implicit = 0.;
    m_volume_around = 0.;
    m_constrained = false;
    m_constraint = 0.;
  }

  My_vertex_base(const Point & p) : Vb(p)
  {
    m_index = -1;
    m_k = -1;
    m_distance = 0.;
    m_implicit = 0.;
    m_volume_around = 0.;
    m_constrained = false;
    m_constraint = 0.;
  }

  My_vertex_base(const Point & p, Cell_handle c) : Vb(p,c)
  {
    m_index = -1;
    m_k = -1;
    m_distance = 0.;
    m_implicit = 0.;
    m_volume_around = 0.;
    m_constrained = false;
    m_constraint = 0.;
  }

  My_vertex_base(Cell_handle c) : Vb(c)
  {
    m_index = -1;
    m_k = -1;
    m_distance = 0.;
    m_implicit = 0.;
    m_volume_around = 0.;
    m_constrained = false;
    m_constraint = 0.;
  }



  const int k () const { return m_k; }
  int& k () { return m_k; }
  const double distance () const { return m_distance; }
  double& distance () { return m_distance; }
  const double implicit () const { return m_implicit; }
  double& implicit () { return m_implicit; }
  const double volume_aounrd () const { return m_volume_around; }
  double& volume_around () { return m_volume_around; }
  const int index () const { return m_index; }
  int& index () { return m_index; }
  const bool constrained () const { return m_constrained; }
  bool& constrained () { return m_constrained; }
  const double constraint () const { return m_constraint; }
  double& constraint () { return m_constraint; }

};



template <class Kernel, class TDS>
class DT3 : public CGAL::Delaunay_triangulation_3<Kernel, TDS>
{
public:

  typedef DT3<Kernel,TDS> Dt;
  typedef typename TDS::Edge Edge;
  typedef typename TDS::Edge Edge_handle;
  typedef typename TDS::Facet Facet_handle;
  typedef typename TDS::Cell_handle Cell_handle;
  typedef typename TDS::Vertex_handle Vertex_handle;
  typedef typename Dt::Finite_vertices_iterator Finite_vertices_iterator;
  typedef typename Dt::All_vertices_iterator All_vertices_iterator;
  typedef typename Dt::Finite_facets_iterator Finite_faces_iterator;
  typedef typename Dt::All_facets_iterator All_facets_iterator;
  typedef typename Dt::Finite_cells_iterator Finite_cells_iterator;
  typedef typename Dt::All_cells_iterator All_cells_iterator;
  typedef typename Dt::Finite_edges_iterator Finite_edges_iterator;
  typedef typename Dt::All_edges_iterator All_edges_iterator;
  typedef typename Dt::Point_iterator Point_iterator;
  typedef typename TDS::Cell_circulator Cell_circulator;
  typedef enum Dt::Locate_type Locate_type;
  typedef typename Kernel::FT FT;
  typedef typename Kernel::Ray_3 Ray;
  typedef typename Kernel::Point_3 Point;
  typedef typename Kernel::Point_2 Point_2;
  typedef typename Kernel::Line_3 Line;
  typedef typename Kernel::Plane_3 Plane;
  typedef typename Kernel::Sphere_3 Sphere;
  typedef typename Kernel::Vector_3 Vector;
  typedef typename Kernel::Segment_3 Segment;
  typedef typename Kernel::Triangle_3 Triangle;
  typedef typename Kernel::Iso_cuboid_3 Iso_cuboid;
  typedef typename Kernel::Tetrahedron_3 Tetrahedron;
  typedef typename CGAL::Polyhedron_3<Kernel> Polyhedron;
  typedef CPCell<FT, Cell_handle> PCell;
  typedef CDPQueue<PCell> DPQueue;
  typedef CGAL::Object Object;
  typedef CNeighbor_search<Kernel> Knn;
  typedef typename std::pair<Triangle,Vector> Facet;
  
private:

  FT m_min_distance;
  FT m_max_distance;
  FT m_min_implicit;
  FT m_max_implicit;

public:

  DT3()
  {
    m_min_distance = 0.;
    m_max_distance = 1.;
    m_min_implicit = -1.;
    m_max_implicit = 1.;
  }
  ~DT3() { }
  

  
  bool is_inside_triangulation (const Point& query)
  {
    Cell_handle c = this->locate (query);
    return !(this->is_infinite (c));
  }



  
  bool close_to_border (Vertex_handle v)
  {
    std::vector<Cell_handle> cn;
    this->incident_cells (v, std::back_inserter (cn));
    for (int i = 0; i < cn.size (); ++ i)
      for (int j = 0; j < 4; ++ j)
	if (on_border (cn[i]->vertex (j)))
	  return true;
    return false;
  }



  bool on_border (Vertex_handle v)
  {
    std::vector<Cell_handle> cn;
    this->incident_cells (v, std::back_inserter (cn));
    for (int i = 0; i < cn.size (); ++ i)
      if (this->is_infinite (cn[i]))
	return true;
    return false;
  }


  void reset_indices ()
  {
    for (Finite_vertices_iterator it = this->finite_vertices_begin ();
	 it != this->finite_vertices_end (); ++ it)
      it->index () = -1;
  }


  void precompute_volumes_around_vertices ()
  {
    std::vector<Vertex_handle> v_vector;

    for(Finite_vertices_iterator v = finite_vertices_begin();
	v != finite_vertices_end(); ++ v)
      v_vector.push_back (v);

#pragma omp parallel for schedule (dynamic)
    for(int v = 0; v < v_vector.size (); ++ v)
      {
	std::vector<Cell_handle> cn;
	this->incident_cells (v_vector[v], std::back_inserter (cn));

	double volume = 0.;

	for (int i = 0; i < cn.size (); ++ i)
	  volume += cn[i]->volume ();

	v_vector[v]->volume_around () =  volume;
      }
  }


  Vertex_handle nearest_vertex (const Point& query)
  {
    Vertex_handle out = NULL;
    Cell_handle c = this->locate (query);    
    double dist_min = std::numeric_limits<double>::max ();
    for (int i = 0; i < 4; ++ i)
      {
	Point p = c->vertex (i)->point ();
	Segment s (query, p);
	double dist = s.squared_length ();
	if (dist < dist_min)
	  {
	    out = c->vertex (i);
	    dist_min = dist;
	  }
      }
    return out;
  }



  Point get_random_point_with_implicit_negative (double value = 0.0)
  {
    while (true)
      {
	int index = rand () % number_of_vertices ();
	Finite_vertices_iterator v;
	int i = 0;
	for(v = finite_vertices_begin();
	    v != finite_vertices_end();
	    ++ v, ++ i)
	  if (i == index)
	    {
	      if (v->implicit () < value)
		return v->point ();
	      break;
	    }
      }
  }



  template <class InputIterator>
  void add_samples_to_triangulation (InputIterator begin,
				     InputIterator end,
				     Knn& knn,
				     double ratio,
				     bool verbose = false)
  {
    if (verbose)
      std::cout << "   Adding " << 100. * ratio << "% of samples in"
		<< " the triangulation.\n";

    for (InputIterator it = begin; it != end; ++ it)
      if (rand () / (double)RAND_MAX < ratio)
	{
	  Point p = *it;
	  Vertex_handle v = this->insert (p);
	  FT distance;
	  unsigned int k;
	  knn.variable_k_function (p, distance, k);
	  v->distance() = distance;
	  v->k () = k;
	  m_max_distance = CGAL::max (m_max_distance, distance);
	  m_min_distance = CGAL::min (m_min_distance, distance);
	}
  }



  double radius_edge_ratio(Cell_handle c)
  {
    double r = circumradius(c);
    double l = len_shortest_edge(c);
    if(l != 0.0)
      return r/l;
    else
      return 1e38;
  }



  FT circumradius(Cell_handle c)
  {
    Point center = this->dual(c);
    const Point& p = c->vertex(0)->point();
    return std::sqrt((p-center)*((p-center)));
  }



  FT distance(const Point& a, const Point& b)
  {
    return std::sqrt(CGAL::squared_distance(a,b));
  }



  double distance(Vertex_handle v1, Vertex_handle v2)
  {
    const Point& a = v1->point();
    const Point& b = v2->point();
    return std::sqrt(CGAL::squared_distance(a,b));
  }



  double len_shortest_edge(Cell_handle c)
  {
    double d1 = distance(c->vertex(0),c->vertex(1));
    double d2 = distance(c->vertex(0),c->vertex(2));
    double d3 = distance(c->vertex(0),c->vertex(3));
    double d4 = distance(c->vertex(1),c->vertex(2));
    double d5 = distance(c->vertex(1),c->vertex(3));
    double d6 = distance(c->vertex(2),c->vertex(3));
    return ::std::min(::std::min(::std::min(d1,d2),d3),
		      ::std::min(::std::min(d4,d5),d6));
  }


  FT max_error_of_cell (Cell_handle c, Knn& knn,
			const unsigned int nb_probes, FT error_limit = 1e30)
  {
    FT max_error = 0.;

    {
      Point p = c->get_center ();
      FT d_real = knn.variable_k_function (p);
      FT d_appr = c->distance_interpolated (p);

      FT error = std::fabs (d_real - d_appr);

      if (error > error_limit) return error;
      if (error > max_error) max_error = error;
    }

    for (int i = 1; i < nb_probes; ++ i)
      {
	Point p = c->get_random_point_inside_cell ();

	FT d_real = knn.variable_k_function (p);
	FT d_appr = c->distance_interpolated (p);

	FT error = std::fabs (d_real - d_appr);

	// test to remove if results are bad
	if (error > error_limit) return error;

	if (error > max_error) max_error = error;

	if (i >= nb_probes / 3 && max_error < 0.5 * error_limit)
	  return max_error;
	//	if (error > error_limit) nb_above ++;
      }

    return max_error;
  }



  unsigned int probe_refinement (Polyhedron& bounding_volume,
				 const FT enlarge_ratio,
				 const unsigned int nb_probes,
				 const FT error_limit,
				 Knn& knn,
				 bool verbose = true)
  {

    m_max_distance = -std::numeric_limits<double>::max ();
    m_min_distance = std::numeric_limits<double>::max ();

    Point centroid = CGAL::centroid (bounding_volume.points_begin (),
				     bounding_volume.points_end ());

    double ratio = 1.;

    if (bounding_volume.size_of_vertices () > 800)
      ratio = 800. / (double)bounding_volume.size_of_vertices ();

    for (typename Polyhedron::Point_iterator it = bounding_volume.points_begin ();
	 it != bounding_volume.points_end (); ++ it)
      if (rand () / (double)RAND_MAX < ratio)
	{
	  Vector vec (centroid, *it);
	  Point p = centroid + enlarge_ratio * vec;

	  Vertex_handle v = this->insert (p);
	  FT distance;
	  unsigned int k;
	  knn.variable_k_function (p, distance, k);
	  v->distance() = distance;
	  v->k () = k;

	  m_max_distance = CGAL::max (m_max_distance, distance);
	  m_min_distance = CGAL::min (m_min_distance, distance);
	}

    if (verbose)
      std::cout << "   Initialised with " << number_of_vertices () << " vertices.\n";

    // Init priority queue
    DPQueue cells;

    Finite_cells_iterator cit;
    for (cit = finite_cells_begin (); cit != finite_cells_end (); ++ cit)
      cells.push (PCell (cit, max_error_of_cell (cit, knn, nb_probes)));


    // Loop
    int nb = 0;
    while (!cells.empty ())
      {
	++ nb;
	// Get a copy of the candidate cell
	PCell pc = cells.top ();
	cells.pop ();
	Cell_handle c = pc.cell ();

	// Get circumcenter and conflict zone
	Point p = c->circumcenter ();

	Locate_type lt;
	int li, lj;
	Cell_handle cloc = this->locate(p, lt, li, lj);
	if (lt == Dt::VERTEX)
	  continue; // Point already exists
	if (this->is_infinite (cloc))
	  continue;


	std::vector<Cell_handle> cells_to_remove;
	this->find_conflicts (p, cloc, CGAL::Emptyset_iterator (),
			      std::back_inserter (cells_to_remove));

	// Remove cells of conflict zone from queue
	for (int i = 0; i < cells_to_remove.size (); ++ i)
	  cells.remove (PCell (cells_to_remove[i]));

	// Add new vertex
	Vertex_handle v = this->insert (p);

	FT distance;
	unsigned int k;
	knn.variable_k_function (p, distance, k);
	v->distance() = distance;
	v->k () = k;

	m_max_distance = CGAL::max (m_max_distance, distance);
	m_min_distance = CGAL::min (m_min_distance, distance);

	// Test neighbor cells and add the ones with error above thres
	std::vector<Cell_handle> cn;
	this->incident_cells (v, std::back_inserter (cn));
	int nb_cells = cn.size ();
	if (nb_cells >= 9)
	  {


#pragma omp parallel for schedule (dynamic)
	    for (int n = 0; n < nb_cells; ++ n)
	      {
		if (radius_edge_ratio (cn[n]) > 1.4)
		  {
#pragma omp critical
		    {
		      cells.push (PCell (cn[n],
					 std::numeric_limits<double>::max ()));
		    }
		    continue;
		  }
		FT error = max_error_of_cell (cn[n], knn, nb_probes,
					      error_limit);
		if (error > error_limit)
		  {
#pragma omp critical
		    {
		      cells.push (PCell (cn[n], error));
		    }
		  }
	      }
	  }
	else
	  {
	    for (int n = 0; n < nb_cells; ++ n)
	      {
		if (radius_edge_ratio (cn[n]) > 1.4)
		  {
		    cells.push (PCell (cn[n],
				       std::numeric_limits<double>::max ()));
		    continue;
		  }

		FT error = max_error_of_cell (cn[n], knn, nb_probes,
					      error_limit);
		if (error > error_limit)
		  cells.push (PCell (cn[n], error));
	      }
	  }

	if (verbose && !(nb % 5000))
	  std::cout << "   Remaining cells to refine: "
		    << cells.size () << std::endl;
      }
    if (verbose)
      std::cout << number_of_vertices () << " vertices created." << std::endl;

  }


  double locate_and_evaluate_distance (const Point& query) const
  {
    Cell_handle c = this->locate (query);
    return c->distance_interpolated (query);
  }

  double locate_and_evaluate_implicit (const Point& query) const
  {
    Cell_handle c = this->locate (query);
    if (this->is_infinite (c))
      return 1000.0;

    return c->implicit_interpolated (query);
  }

  bool locate_and_evaluate_implicit (const Point& query,
				     double& value) const
  {
    Cell_handle c = this->locate (query);
    if (this->is_infinite (c))
      return false;
    value = c->implicit_interpolated (query);
    return true;
  }

  Cell_handle smart_locate (const Point& query, Cell_handle prev)
  {
    if (prev != NULL && prev->contains (query))
      return prev;

    return this->locate (query, prev);
  }

  FT weighted_laplacian_coef (Edge& edge, unsigned int kmax)
  {
    Cell_handle c = edge.first;

    Vertex_handle v1 = c->vertex (edge.second);
    Vertex_handle v2 = c->vertex (edge.third);

    FT dist1 = v1->k () < kmax ? v1->distance () : 100;
    FT dist2 = v2->k () < kmax ? v2->distance () : 100;
    FT dist = CGAL::min (dist1, dist2);

    return dist * laplacian_dec_coefficient (edge);
  }


  // cotan formula as area(voronoi face) / len(primal edge)
  FT laplacian_dec_coefficient(Edge& edge)
  {
    Cell_handle cell = edge.first;
    Vertex_handle vi = cell->vertex(edge.second);
    Vertex_handle vj = cell->vertex(edge.third);

    // primal edge
    const Point& pi = vi->point();
    const Point& pj = vj->point();
    Vector primal = pj - pi;
    FT len_primal = std::sqrt(primal * primal);

    return area_voronoi_face(edge) / len_primal;
  }

  // spin around edge
  // return area(voronoi face)
  FT area_voronoi_face(Edge& edge)
  {
    // circulate around edge
    Cell_circulator circ = this->incident_cells(edge);
    Cell_circulator done = circ;
    std::vector<Point> voronoi_points;
    do
      {
	Cell_handle cell = circ;
	if(!this->is_infinite(cell))
	  voronoi_points.push_back(this->dual(cell));
	else // one infinite tet, switch to another calculation
	  return area_voronoi_face_boundary(edge);
	circ++;
      }
    while(circ != done);

    if(voronoi_points.size() < 3)
      return 0.0;

    // sum up areas
    FT area = 0.0;
    const Point& a = voronoi_points[0];
    unsigned int nb_triangles = voronoi_points.size() - 1;
    for(unsigned int i=1;i<nb_triangles;i++)
      {
	const Point& b = voronoi_points[i];
	const Point& c = voronoi_points[i+1];
	Triangle triangle(a,b,c);
	area += std::sqrt(triangle.squared_area());
      }

    return area;
  }

  // approximate area when a cell is infinite
  FT area_voronoi_face_boundary(Edge& edge)
  {
    FT area = 0.0;
    Vertex_handle vi = edge.first->vertex(edge.second);
    Vertex_handle vj = edge.first->vertex(edge.third);

    const Point& pi = vi->point();
    const Point& pj = vj->point();
    Point m = CGAL::midpoint(pi,pj);

    // circulate around each incident cell
    Cell_circulator circ = this->incident_cells(edge);
    Cell_circulator done = circ;
    do
      {
	Cell_handle cell = circ;
	if(!this->is_infinite(cell))
	  {
	    // circumcenter of cell
	    Point c = this->dual(cell);
	    Tetrahedron tet = this->tetrahedron(cell);

	    int i = cell->index(vi);
	    int j = cell->index(vj);
	    int k = -1, l = -1;
	    other_two_indices(i,j, k,l);
	    Vertex_handle vk = cell->vertex(k);
	    Vertex_handle vl = cell->vertex(l);

	    const Point& pk = vk->point();
	    const Point& pl = vl->point();

	    // if circumcenter is outside tet
	    // pick barycenter instead
	    if(tet.has_on_unbounded_side(c))
	      c = CGAL::centroid(pi,pj,pk,pl);

	    Point ck = CGAL::circumcenter(pi,pj,pk);
	    Point cl = CGAL::circumcenter(pi,pj,pl);

	    Triangle mcck(m,m,ck);
	    Triangle mccl(m,m,cl);

	    area += std::sqrt(mcck.squared_area());
	    area += std::sqrt(mccl.squared_area());
	  }
	circ++;
      }
    while(circ != done);
    return area;
  }

  // Gets indices different from i and j
  void other_two_indices(int i,
			 int j,
			 int& k,
			 int& l)
  {
    bool k_done = false;
    bool l_done = false;
    for(int index=0;index<4;index++)
      {
	if(index != i && index != j)
	  {
	    if(!k_done)
	      {
		k = index;
		k_done = true;
	      }
	    else
	      {
		l = index;
		l_done = true;
	      }
	  }
      }
  }


  Vertex_handle vertex (const Point& p)
  {
    Vertex_handle out;
    if (!(this->is_vertex (p, out)))
      Utils::error ("Point is not a vertex of triangulation.");
    return out;
  }

  void switch_signs_if_opposite (bool verbose)
  {
    // Check sign of border domain
    std::vector<Vertex_handle> border;
    this->incident_vertices (this->infinite_vertex (),
			     std::back_inserter (border));
    int nb;
    for (unsigned int i = 0; i < border.size (); ++ i)
      if (border[i]->implicit () > 0.)
	++ nb;

    // If sign is positive, it's okay
    if (2 * nb > border.size ())
      return;

    Utils::print (verbose, "   Switching implicit function signs...");

    // If sign is negative, switch all the signs
    for (Finite_vertices_iterator it = this->finite_vertices_begin ();
	 it != this->finite_vertices_end (); ++ it)
      it->implicit () = -it->implicit ();


  }


  template <class InputIterator>
  double median_implicit_at_input_points (InputIterator begin,
					  InputIterator end)
  {
    // compute all values
    InputIterator it;
    std::vector<double> values;
    for(it = begin; it != end; it++)
      values.push_back (locate_and_evaluate_implicit (*it));

    // sort them in increasing order
    std::sort(values.begin(),values.end());
    return values[values.size()/2];
  }


};

#endif // DT__
