/**
 * NASR - Noise-Adaptive Shape Reconstruction
 * Source code, Version 1.0 (December 2014)
 *
 * Copyright (c) 2014 Simon Giraudot <simon.giraudot@inria.fr>
 * http://www-sop.inria.fr/members/Simon.Giraudot/
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#ifndef __GRAPH__
#define __GRAPH__

#define EIGEN_YES_I_KNOW_SPARSE_MODULE_IS_NOT_STABLE_YET
#include <eigen3/Eigen/Sparse>

template <typename Kernel>
class CNode
{
public :

  typedef typename Kernel::FT FT;
  typedef typename Kernel::Point_3 Point;
  typedef typename Kernel::Vector_3 Vector;

private :

  Point m_point;
  std::map <int, double> m_coefs;
  double m_diag_coef;
  FT m_sign;
  FT m_conf;

public :

  CNode () : m_point (Point (0., 0., 0.)), m_sign (0.), m_conf (1.) { }
  CNode (const Point& p) : m_point (p), m_sign (0.), m_conf (1.) { }
  CNode (const Point& p, const FT& s) : m_point (p), m_sign (s), m_conf (1.) { }

  const Point& point () const { return m_point; }
  Point& point () { return m_point; }

  std::map <int, double>& coefs () { return m_coefs; }
  double diag_coef () { return m_diag_coef; }
  double sign () { return m_sign; }
  double conf () const { return m_conf; }

  void add_coef (int i, double coef)
  {
    m_coefs.insert (std::pair<int, double> (i, coef)).second;
    m_diag_coef += std::fabs (coef);
  }

  void set_sign (double sign)
  {
    m_sign = sign;
  }

  void set_conf (double conf)
  {
    m_conf = conf;
  }

};

template <typename Kernel, typename Dt3>
class CGraph
{
private:

  typedef CNode<Kernel> Node;
  typedef std::vector<Node> Nodes;
  typedef typename Dt3::Vertex_handle Vertex_handle;
  typedef typename Dt3::Finite_vertices_iterator Finite_vertices_iterator;
  typedef typename Dt3::Edge Edge;
  typedef typename Dt3::Cell_handle Cell_handle;

  Dt3& m_dt3;
  Nodes m_nodes;

public:

  CGraph (Dt3& dt3) : m_dt3 (dt3) { }
  virtual ~CGraph () { }

  void clear () { m_nodes.clear (); }

  unsigned int size () const { return m_nodes.size (); }
  Node& operator[] (const int index) { return m_nodes[index]; }

  void build_uniform (const Iso_cuboid& bounding_box,
		      const int nodes)
  {
    double size = 1.3;
    double dx = size * (bounding_box.xmax () - bounding_box.xmin ());
    double dy = size * (bounding_box.ymax () - bounding_box.ymin ());
    double dz = size * (bounding_box.zmax () - bounding_box.zmin ());

    double x_min = - dx / 2;
    double y_min = - dy / 2;
    double z_min = - dz / 2;

    int nb_x = (int)(Utils::cbrt (dx * dx * nodes / (dy * dz)));
    int nb_y = (int)(Utils::cbrt (dy * dy * nodes / (dx * dz)));
    int nb_z = (int)(Utils::cbrt (dz * dz * nodes / (dx * dy)));

    for (int i = 0; i < nb_x; ++ i)
      for (int j = 0; j < nb_y; ++ j)
	for (int k = 0; k < nb_z; ++ k)
	  {
	    Point p (x_min + dx * i / (double)(nb_x-1),
		     y_min + dy * j / (double)(nb_y-1),
		     z_min + dz * k / (double)(nb_z-1));

	    if (!(m_dt3.is_inside_triangulation (p)))
	      continue;

	    m_nodes.push_back (Node (p));
	  }
  }

  void build_on_dt ()
  {
    m_dt3.reset_indices ();

    int count = 0;
    for (Finite_vertices_iterator it = m_dt3.finite_vertices_begin ();
	 it != m_dt3.finite_vertices_end (); ++ it)
      {
	m_nodes.push_back (Node (it->point ()));
	it->index () = count ++;
      }
  }

  void pick_random_edges (const int edges, const unsigned int kmax, bool verbose)
  {
    double step = 0.01; // Discretizing step for an edge
    double dist_min = Utils::square (15. * step); // 15 steps, then a sheer drop
    double dist_max = 1.; // Half the bounding box

    if (dist_max < 2. * dist_min)
      dist_max = 2. * dist_min;

    if (verbose)
      std::cout << "   Edges picked with length in [ " << dist_min
		<< " ; " << dist_max << "]\n";

    time_t start = time (NULL);
    time_t last_refresh = start;
#pragma omp parallel for schedule (dynamic)
    for (int i = 0; i < m_nodes.size (); ++ i)
      {
	std::set<int> already_done;
	already_done.insert (i);
	Point a = m_nodes[i].point ();

	Cell_handle cell_located = m_dt3.locate (a);

	for (int j = 0; j < edges; ++ j)
	  {
	    int index = 0;
	    bool ok = false;
	    Point b;
	    double dist = 0.;

	    while (!ok)
	      {
		index = rand () % (m_nodes.size ());

		if (already_done.insert (index).second == false)
		  continue;

		b = m_nodes[index].point ();

		dist = CGAL::squared_distance (a, b);

		if (dist < dist_min || dist > dist_max)
		  continue;

		ok = true;
	      }

	    int nb_peaks = detect_peaks (a, b, cell_located, step, kmax);
	    int coef = (nb_peaks % 2) ? 1 : -1;

#pragma omp critical
	    {
	      m_nodes[i].add_coef (index, coef);
	      m_nodes[index].add_coef (i, coef);
	    }
	  }
#pragma omp critical
	  {
	    if (verbose
		&& time (NULL) - last_refresh > 1) // Display progress
	      {
		Utils::display_progress (i, m_nodes.size (), start);
		last_refresh = time (NULL);
	      }
	  }
      }
    if (verbose) std::cout << std::endl;
  }

  void solve_sign_guess (bool verbose)
  {
    int nb_nodes = m_nodes.size ();

    // Eigen part

    // To fill a sparse matrix, Eigen uses a set of triplets (2 integers
    // for the indices + a templated value). If you want to use a dense
    // matrix, it's probably different.
    typedef Eigen::Triplet<double> T;
    std::vector<T> cells;

    // Create the matrix and vector
    Eigen::SparseMatrix<double> A (nb_nodes, nb_nodes);
    Eigen::SparseMatrix<double> C (nb_nodes, nb_nodes);
    Eigen::VectorXd b (nb_nodes);

    // Fill matrix and vector
    for (int i = 0; i < nb_nodes; ++ i)
      {
	cells.push_back (T (i, i, m_nodes[i].diag_coef ()));

	std::map<int, double>& coefs = m_nodes[i].coefs ();
	std::map<int, double>::iterator it;

	for (it = coefs.begin (); it != coefs.end (); ++ it)
	  cells.push_back (T (i, it->first, it->second));

	b(i) = 1.;
      }
    A.setFromTriplets (cells.begin (), cells.end ());

    if (verbose)
      {
	std::cout << "   Overview of the matrix A and vector B:\n";

	for (int i = 0; i < nb_nodes;
	     i += nb_nodes / 10)
	  {
	    std::cout << "     ";
	    for (int j = 0; j < nb_nodes;
		 j += nb_nodes / 10)
	      { 
		std::cout << A.coeff(i,j) << " ";
		if (i == j && A.coeff(i,j) < 10)
		  std::cout << " ";
	      }
	    std::cout << "   " << b(i) << std::endl;
	  }
      }

    Eigen::VectorXd res;

    Eigen::ConjugateGradient<Eigen::SparseMatrix<double> > cg;
    cg.compute (A);
    res = cg.solve (b);
  

    bool surface = false;

    for (int i = 0; i < nb_nodes; ++ i)
      {
	m_nodes[i].set_sign ((res(i) > 0.) ? 1. : -1.);
	if (res (i) < 0.)
	  surface = true;
      }

    double mean_conf = 0;
    for (int i = 0; i < nb_nodes; ++ i)
      {
	double conf = compute_conf (i);
	mean_conf += conf;
	m_nodes[i].set_conf (conf);
      }

    if (verbose)
      {
	std::cout << "   Mean confidence on node signs: "
		  << mean_conf / nb_nodes << std::endl;
      }

    if (!surface)
      Utils::error ("All nodes are positive, no inside found.");

  }



  void solve_random_walker (const double alpha, const unsigned int kmax,
			    bool verbose)
  {
    int nb_nodes = m_nodes.size ();

    Eigen::SparseMatrix<double> A (nb_nodes, nb_nodes);
    Eigen::VectorXd b (nb_nodes);

    typedef Eigen::Triplet<double> T;
    std::vector<T> A_cells;

    m_dt3.precompute_volumes_around_vertices ();

    double mean_diag = 0.;
    time_t start = time (NULL);
    time_t last_refresh = start;

    std::vector<double> diagonal;

    for (int i = 0; i < nb_nodes; ++ i)
      {
	Vertex_handle v = m_dt3.vertex (m_nodes[i].point ());

	std::vector<Edge> edges;
	m_dt3.finite_incident_edges (v, std::back_inserter (edges));

	double volume = v->volume_around ();

	diagonal.push_back (0.);

	for (int e = 0; e < edges.size (); ++ e)
	  {
	    Edge edge = edges[e];

	    Vertex_handle vj;

	    if(edge.first->vertex(edge.second) == v)
	      vj = edge.first->vertex(edge.third);
	    else
	      vj = edge.first->vertex(edge.second);

	    int index = vj->index ();

	    double cij = 0.5 * (vj->volume_around () + volume) *
	      m_dt3.weighted_laplacian_coef (edge, kmax);

	    if (isnan (cij) || cij == std::numeric_limits<double>::infinity ()
		|| cij == -std::numeric_limits<double>::infinity ())
	      continue;

	    diagonal.back () += cij;
	    A_cells.push_back (T (i, index, -cij));
	  }

	mean_diag += diagonal.back ();

	if (verbose
	    && time (NULL) - last_refresh > 1) // Display progress
	  {
	    Utils::display_progress (i, nb_nodes, start);
	    last_refresh = time (NULL);
	  }
      }


    if (verbose) std::cout << std::endl;

    mean_diag /= nb_nodes;

    if (verbose)
      std::cout << "Mean value of the diagonal coefficients: " <<
	mean_diag << std::endl;

    double weight = alpha * mean_diag;

    for (int i = 0; i < nb_nodes; ++ i)
      {
	Vertex_handle v = m_dt3.vertex (m_nodes[i].point ());

	b(i) = 0.;
	if (v->constrained ())
	  {
	    diagonal[i] += weight;
	    b(i) = weight * v->constraint ();
	  }
	A_cells.push_back (T (i, i, diagonal[i]));
      }

    A.setFromTriplets (A_cells.begin (), A_cells.end ());


    if (verbose)
      {
	std::cout << "   Overview of the matrix A and vector B:\n";

	for (int i = 0; i < nb_nodes;
	     i += nb_nodes / 10)
	  {
	    std::cout << "     ";
	    for (int j = 0; j < nb_nodes;
		 j += nb_nodes / 10)
	      std::cout << A.coeff(i,j) << " ";
	    std::cout << "   |   " << b(i) << std::endl;
	  }
      }

    Eigen::ConjugateGradient<Eigen::SparseMatrix<double> > cg;
    Eigen::VectorXd x (nb_nodes);
    cg.setMaxIterations (50000);
    cg.compute (A);
    x = cg.solve (b);

    if (verbose)
      std::cout << "   Iterations: " << cg.iterations() << std::endl
		<< "   Error: " << cg.error() << std::endl;

    bool surface = false;

    for (Finite_vertices_iterator it = m_dt3.finite_vertices_begin ();
	 it != m_dt3.finite_vertices_end (); ++ it)
      if (it->index () != 1)
	{
	  it->implicit () = x(it->index ());
	  if (it->implicit () < 0.)
	    surface = true;
	}

    if (!surface)
      Utils::error ("Implicit function is positive, no surface found.");
  }



  int detect_peaks (const Point& a, const Point& b, Cell_handle c,
		    const double step, const unsigned int kmax)
  {
    typedef std::vector<double> vec_double;

    double length = std::sqrt (CGAL::squared_distance (a, b));
    unsigned int nb_points = (unsigned int)(length / step);

    vec_double distances;
    std::vector<int> k;
    compute_straight_segment (a, b,
			      nb_points,
			      std::back_inserter (distances),
			      std::back_inserter (k),
			      c);

    int size = distances.size ();

    // Find local minima
    std::vector<int> indices_minima;

    vec_double smoothed;
    light_smooth (distances, std::back_inserter (smoothed));

    for (int i = 0; i < size; ++ i)
      {
	if (smoothed[i] < smoothed[i-1] &&
	    smoothed[i] < smoothed[i+1])
	  {
	    if (k[i] < kmax)
	      indices_minima.push_back (i);
	  }
      }

    int combi = 1;
    for (int i = 0; i < indices_minima.size (); ++ i)
      combi *= 2;

    double best = std::numeric_limits<double>::max ();
    int best_combi = 0;
    int best_nb = 0;
    for (int i = 0; i < combi; ++ i)
      {
	std::vector<int> minima;
	for (int m = 0; m < indices_minima.size (); ++ m)
	  {
	    bool set = Utils::is_bit_set (i, m);
	    if (set)
	      minima.push_back (indices_minima[m]);
	  }
	vec_double flipped;
	local_multiflip_function (distances, minima,
				  std::back_inserter (flipped));
	vec_double fsmoothed;
	light_smooth (flipped, std::back_inserter (fsmoothed));

	double norm = quick_full_regularity_norm (fsmoothed);

	if (norm < best)
	  {
	    best = norm;
	    best_combi = i;
	    best_nb = minima.size ();
	  }
      }

    return best_nb;
  }


  template <typename Iterator1, typename Iterator2>
  void compute_straight_segment (const Point& source,
				 const Point& target,
				 const unsigned int nb_points,
				 Iterator1 distances,
				 Iterator2 k,
				 Cell_handle c = NULL)
  {
    Vector direction (source, target);

    Line line (source, direction);

    direction = direction / std::sqrt (direction.squared_length ());

    if (c == NULL)
      c = m_dt3.locate (source);

    std::ofstream f ("test.plot");
    for (int i = 0; i < nb_points; ++ i)
      {
	double coef = i / (double)nb_points;
	Point p (source.x () * (1. - coef) + target.x () * coef,
		 source.y () * (1. - coef) + target.y () * coef,
		 source.z () * (1. - coef) + target.z () * coef);

	c = m_dt3.smart_locate (p, c);
	double dist;
	unsigned int kk;
	c->distance_and_k (p, dist, kk);
	f << dist << " " << kk << std::endl;
	*(distances ++) = dist;
	*(k ++) = kk;
      }
    f.close ();
  }

  template <typename Iterator>
  void light_smooth (const std::vector<double>& f, Iterator out)
  {
    const int size = f.size () - 2;
    *(out ++) = (6. * f[0] + 4. * f[1] + f[2]) / 11.;
    *(out ++) = (4. * f[0] + 6. * f[1] + 4. * f[2] + f[3]) / 15.;

    for (int i = 2; i < size; ++ i)
      *(out ++) = (f[i - 2] +
		   4. * f[i - 1] +
		   6. * f[i] +
		   4. * f[i + 1] +
		   f[i + 2]) / 16.;

    *(out ++) = (f[f.size () - 4] +
		 4. * f[f.size () - 3] +
		 6. * f[f.size () - 2] +
		 4. * f[f.size () - 1]) / 15.;
    *(out ++) = (f[f.size () - 3] +
		 4. * f[f.size () - 2] +
		 6. * f[f.size () - 1]) / 11.;
  }

  template <typename Iterator>
  void local_multiflip_function (const std::vector<double>& f,
				 std::vector<int>& indices,
				 Iterator out)
  {
    int sign = 1;
    int current_index = 0;
    double offset = 0;
    for (int i = 0; i < f.size (); ++ i)
      {
	double F = offset + f[i] * sign;
	if (current_index < indices.size () && i == indices[current_index])
	  {
	    offset = F + f[i] * sign;
	    sign = -sign;
	    *(out ++) = F;
	    ++ current_index;
	    continue;
	  }
	*(out ++) = F;
      }
  }

  double quick_full_regularity_norm  (const std::vector<double>& f)
  {
    const unsigned int step = 7;
    double out = 0.;
    int size = f.size ();

    double* ddf[step];
    for (int s = 1; s <= step; ++ s)
      {
	ddf[s-1] = new double[size + 2 * s];
	for (int i = 0; i < size + 2 * s; ++ i)
	  ddf[s-1][i] = 0.;
      }

    for (int i = 0; i < size; ++ i)
      {
	double val = f[i];
	for (int s = 1; s <= step; ++ s)
	  {
	    ddf[s-1][i] += val;
	    ddf[s-1][i+s] -= 2*val;
	    ddf[s-1][i+2*s] += val;
	  }
      }

    for (int s = 1; s <= step; ++ s)
      {
	double norm = 0.;
	for (int i = 2 * s; i < size; ++ i)
	  norm += Utils::square (ddf[s-1][i]);
	out += std::sqrt (norm) / (s*s);
      }

    for (int s = 1; s <= step; ++ s)
      delete[] ddf[s-1];

    return out / step;
  }


  double compute_conf (int index)
  {
    std::map<int, double>& coefs = m_nodes[index].coefs ();
    std::map<int, double>::iterator it;
    int sign_a = (m_nodes[index].sign () > 0) ? 1 : -1;

    int nb = 0;
    int nb_satisfied = 0;
    for (it = coefs.begin (); it != coefs.end (); ++ it, ++ nb)
      {
	int sign_b = (m_nodes[it->first].sign () > 0) ? 1 : -1;
	bool same_side = (sign_a * sign_b > 0);
	if (same_side && it->second < 0)
	  nb_satisfied ++;
	if (!same_side && it->second > 0)
	  nb_satisfied ++;
      }

    return (double)nb_satisfied / (double)nb;
  }

  void info ()
  {
    std::cout << "   Graph has " << m_nodes.size () << " nodes" << std::endl;
    unsigned int nb = 0;
    double conf = 0;
    for (unsigned int i = 0; i < m_nodes.size (); ++ i)
      {
	if (m_nodes[i].sign () > 0.)
	  nb ++;
	conf += m_nodes[i].conf ();
      }
    conf /= m_nodes.size ();

    std::cout << "   " << nb << " nodes with positive sign" << std::endl;
    std::cout << "   " << m_nodes.size () - nb
	      << " nodes with negative sign" << std::endl;
    std::cout << "   Average confidence = " << conf << std::endl;
  }

  

};



#endif // GRAPH__
