/**
 * NASR - Noise-Adaptive Shape Reconstruction
 * Source code, Version 1.0 (December 2014)
 *
 * Copyright (c) 2014 Simon Giraudot <simon.giraudot@inria.fr>
 * http://www-sop.inria.fr/members/Simon.Giraudot/
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#ifndef __MESH__
#define __MESH__

#define EIGEN_YES_I_KNOW_SPARSE_MODULE_IS_NOT_STABLE_YET
#include <eigen3/Eigen/Sparse>

#include <CGAL/Surface_mesh_default_triangulation_3.h>
#include <CGAL/Complex_2_in_triangulation_3.h>
#include <CGAL/make_surface_mesh.h>
#include <CGAL/Implicit_surface_3.h>


template <typename Kernel, typename Dt3>
class Zero_function 
{
public:
  typedef CGAL::Bbox_3 BBox;
  typedef Kernel Geom_traits;
  typedef typename Kernel::FT FT;
  typedef typename Kernel::Point_3 Point;
  typedef typename Dt3::Cell_handle Cell_handle;
  typedef Kernel Surface_mesher_traits_3;

private:
  Dt3& m_dt3;
  double m_value;

public:
  Zero_function(Dt3& dt3, const double value)
    : m_dt3(dt3),
      m_value(value)
  {
  }
  ~Zero_function(){}

public:

  FT operator()(const Point& query) const
  {
    double value;
    bool ok = m_dt3.locate_and_evaluate_implicit (query, value);
    if (ok)
      return (value - m_value);
    else
      return 1000.0;
  }

};


template <typename Kernel, typename Dt3>
class CMesh
{
private:

  typedef Zero_function<Kernel, Dt3> Zero_function;

  typedef typename Dt3::Vertex_handle Vertex_handle;
  typedef typename Dt3::Finite_vertices_iterator Finite_vertices_iterator;
  typedef typename Dt3::Edge Edge;
  typedef typename Dt3::Cell_handle Cell_handle;

  typedef CGAL::Surface_mesh_default_triangulation_3 STr;
  typedef CGAL::Complex_2_in_triangulation_3<STr> C2t3;
  typedef CGAL::Implicit_surface_3<Kernel, Zero_function> Surface_3;
  typedef CGAL::Surface_mesh_default_criteria_3<STr> Surface_mesh_criteria;


  Dt3& m_dt3;
  STr m_str;
  C2t3 m_c2t3;

public:

  CMesh (Dt3& dt3) : m_dt3 (dt3), m_str (), m_c2t3 (m_str) { }

  virtual ~CMesh () { }

  void generate (double median_value, const Sphere& s,
		 double mesh_angle, double mesh_radius, double mesh_distance)

  {
    Zero_function surface (m_dt3, median_value);
    STr str;

    Surface_3 domain (surface, s, 1e-6);
    Surface_mesh_criteria criteria (mesh_angle, mesh_radius, mesh_distance);

    CGAL::make_surface_mesh (m_c2t3,
			     domain,
			     criteria,
			     CGAL::Non_manifold_tag (),
			     500);
  }

  const C2t3& get () const { return m_c2t3; }

};



#endif // MESH__
