/**
 * NASR - Noise-Adaptive Shape Reconstruction
 * Source code, Version 1.0 (December 2014)
 *
 * Copyright (c) 2014 Simon Giraudot <simon.giraudot@inria.fr>
 * http://www-sop.inria.fr/members/Simon.Giraudot/
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#ifndef __NEIGHBOR_SEARCH__
#define __NEIGHBOR_SEARCH__

#define PI 3.141592653589793238462643
#define SQRT_3 1.73205081

#include <CGAL/basic.h>
#include <CGAL/eigen.h>
#include <CGAL/Orthogonal_k_neighbor_search.h>
#include <CGAL/Search_traits_3.h>

#include <vector>
#include <list>
#include <cmath>

#include "hierarchical_clustering.h"

template <class Kernel>
class CNeighbor_search
{
public:
  typedef typename Kernel::FT FT;
  typedef typename Kernel::Point_3 Point_3;
  typedef typename Kernel::Plane_3 Plane;
  typedef typename Kernel::Vector_3 Vector;
  typedef typename CGAL::Search_traits_3<Kernel> TreeTraits;
  typedef typename CGAL::Orthogonal_k_neighbor_search<TreeTraits> Neighbor_search;
  typedef typename Neighbor_search::Tree Tree;
  typedef typename Neighbor_search::iterator Iterator;

private:

  int m_cluster_size;
  int m_nb_trees;
  std::vector<Tree*> m_tree;
  int m_nb_points;

public:
  CNeighbor_search()
  {
    m_nb_points = 0;
    m_cluster_size = 10;
  }

  ~CNeighbor_search() {}

  template <class InputIterator>
  void init(InputIterator begin,
	    InputIterator end,
	    const int nb_points)
  {
    for (int i = 0; i < m_tree.size (); ++ i)
      delete m_tree[i];
    m_tree.clear ();

    m_nb_points = nb_points;
    m_nb_trees = 0;

    while (m_nb_points > m_cluster_size)
      {
    	m_nb_trees ++;
    	m_nb_points /= m_cluster_size;
      }

    m_nb_points = nb_points;

    std::vector<Point_3>* point_sets = new std::vector<Point_3>[m_nb_trees];

    typedef typename std::vector<Point_3>::iterator Point_iterator;

    std::copy (begin, end, std::back_inserter (point_sets[0]));
    int cluster_size = m_cluster_size;
    for (int i = 1; i < m_nb_trees; ++ i)
      {
	std::copy (begin, end, std::back_inserter (point_sets[i]));

	Point_iterator removed
	  = CGAL::hierarchical_clustering<Point_iterator, Kernel>
	  (point_sets[i].begin (), point_sets[i].end (),
	   cluster_size, 0.33);
	point_sets[i].erase (removed, point_sets[i].end ());
	cluster_size *= m_cluster_size;
      }

    for (int i = 0; i < m_nb_trees; ++ i)
      m_tree.push_back (new Tree (point_sets[i].begin (), point_sets[i].end ()));

  }

  void internal_variable_k_function(const Point_3& query, FT& result,
				    unsigned int& best_k)
  {
    FT d_min = std::numeric_limits<double>::max ();

    int weight = 1;
    int cluster_size = m_cluster_size;
    FT sum_sq_distances = 0.0;

    for (int t = 0; t < m_nb_trees; ++ t)
      {
	if (t == m_nb_trees - 1) cluster_size *= cluster_size;

	Neighbor_search search(*(m_tree[t]), query, cluster_size);
	Iterator it = search.begin();
	if (t != 0) ++ it;

	int nb = 0;

	for(unsigned int i = 0; i < cluster_size; ++ i, ++ it)
	  {
	    if(it == search.end())
	      break;
	    Point_3 p = get_point (it);
	    sum_sq_distances += weight * CGAL::squared_distance(query,p);
	    nb += weight;

	    if (nb < 6) continue;

	    FT d = std::sqrt (sum_sq_distances / nb)
	      / std::pow (nb, 0.375); // NB ^ (5/12)

	    if (d < d_min)
	      {
		d_min = d;
		best_k = nb;
	      }
	  }
	weight *= m_cluster_size;
      }
    result = d_min;
  }

  void variable_k_function(const Point_3& query,
			   FT& du, unsigned int& k)
  {
    internal_variable_k_function (query, du, k);
  }

  FT variable_k_function(const Point_3& query)
  {
    FT du;
    unsigned int k;
    internal_variable_k_function (query, du, k);
    return du;
  }

  unsigned int k_function(const Point_3& query)
  {
    FT du;
    unsigned int k;
    internal_variable_k_function (query, du, k);
    return k;
  }

  Point_3 get_point(Iterator it)
  {
    return it->first;
  }

};

#endif // _MS_KNN_
