/**
 * NASR - Noise-Adaptive Shape Reconstruction
 * Source code, Version 1.0 (December 2014)
 *
 * Copyright (c) 2014 Simon Giraudot <simon.giraudot@inria.fr>
 * http://www-sop.inria.fr/members/Simon.Giraudot/
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __PARAMETERS__
#define __PARAMETERS__

#include <iostream>
#include <string>

#include <cstdio>
#include <cstring>

#include "common.h"

class Parameters
{
private:

  bool m_verbose;  
  std::string m_input_filename;
  std::string m_output_filename;
  unsigned int m_probes;
  double m_error;
  unsigned int m_nodes;
  unsigned int m_edges;
  double m_conf;
  unsigned int m_kmax;
  double m_alpha;
  double m_mesh_angle;
  double m_mesh_radius;
  double m_mesh_distance;

public:

  Parameters (int argc, char** argv) : m_output_filename ("out.off")
  {
    m_verbose = false;
    m_probes = 10;
    m_error = 0.004;
    m_nodes = 50000;
    m_edges = 30; // Edges per node, 1500000/50000
    m_conf = 0.75;
    m_kmax = 500;
    m_alpha = 1e-3;
    m_mesh_angle = 30.;
    m_mesh_radius = 0.05;
    m_mesh_distance = 0.05;

    if (argc < 2)
      {
	Utils::print ("USAGE: ./nasr [INPUT] [OPTIONS] (see --help)");
	exit (EXIT_FAILURE);
      }

    unsigned int i, p;
    for (i = 1; i < argc; ++ i)
      {
	if (is_option (argv[i], "--help", "-h"))
	  {
	    print_help ();
	    exit (EXIT_SUCCESS);
	  }
	else if (is_option (argv[i], "--output", "-o"))
	  {
	    ++ i;
	    if (i == argc)
	      Utils::error ("Bad number of arguments.");
	    m_output_filename = std::string (argv[i]);
	  }
	else if (is_option (argv[i], "--probes", "-p"))
	  {
	    ++ i;
	    if (i == argc)
	      Utils::error ("Bad number of arguments.");

	    if (sscanf (argv[i], "%u", &m_probes) != 1)
	      Utils::error ("Wrong argument");
	  }
	else if (is_option (argv[i], "--error", "-e"))
	  {
	    ++ i;
	    if (i == argc)
	      Utils::error ("Bad number of arguments.");

	    if (sscanf (argv[i], "%lf", &m_error) != 1)
	      Utils::error ("Wrong argument");
	  }
	else if (is_option (argv[i], "--nodes", "-n"))
	  {
	    ++ i;
	    if (i == argc)
	      Utils::error ("Bad number of arguments.");

	    if (sscanf (argv[i], "%u", &m_nodes) != 1)
	      Utils::error ("Wrong argument");
	  }
	else if (is_option (argv[i], "--edges", "-E"))
	  {
	    ++ i;
	    if (i == argc)
	      Utils::error ("Bad number of arguments.");

	    if (sscanf (argv[i], "%u", &m_edges) != 1)
	      Utils::error ("Wrong argument");
	  }
	else if (is_option (argv[i], "--conf", "-c"))
	  {
	    ++ i;
	    if (i == argc)
	      Utils::error ("Bad number of arguments.");

	    if (sscanf (argv[i], "%lf", &m_conf) != 1)
	      Utils::error ("Wrong argument");
	  }
	else if (is_option (argv[i], "--kmax", "-k"))
	  {
	    ++ i;
	    if (i == argc)
	      Utils::error ("Bad number of arguments.");

	    if (sscanf (argv[i], "%u", &m_kmax) != 1)
	      Utils::error ("Wrong argument");
	  }
	else if (is_option (argv[i], "--alpha", "-a"))
	  {
	    ++ i;
	    if (i == argc)
	      Utils::error ("Bad number of arguments.");

	    if (sscanf (argv[i], "%lf", &m_alpha) != 1)
	      Utils::error ("Wrong argument");
	  }
	else if (is_option (argv[i], "--angle", "-A"))
	  {
	    ++ i;
	    if (i == argc)
	      Utils::error ("Bad number of arguments.");

	    if (sscanf (argv[i], "%lf", &m_mesh_angle) != 1)
	      Utils::error ("Wrong argument");
	  }
	else if (is_option (argv[i], "--radius", "-r"))
	  {
	    ++ i;
	    if (i == argc)
	      Utils::error ("Bad number of arguments.");

	    if (sscanf (argv[i], "%lf", &m_mesh_radius) != 1)
	      Utils::error ("Wrong argument");
	  }
	else if (is_option (argv[i], "--distance", "-D"))
	  {
	    ++ i;
	    if (i == argc)
	      Utils::error ("Bad number of arguments.");

	    if (sscanf (argv[i], "%lf", &m_mesh_distance) != 1)
	      Utils::error ("Wrong argument");
	  }
	else if (is_option (argv[i], "--verbose", "-v"))
	  m_verbose = true;
	else
	  m_input_filename = std::string (argv[i]);
      }

    if (m_input_filename == std::string ())
      Utils::error ("No input file specified.");

    if (m_verbose)
      display_parameters ();

  }

  virtual ~Parameters () { }


  inline bool is_option (char* arg, char* option1, char* option2) const
  {
    return (!strcmp (arg, option1) || !strcmp (arg, option2));
  }

  void print_help () const
  {
    Utils::print ("NASR (Noise-Adaptive Shape Reconstruction)");
    Utils::print ("");
    Utils::print ("./nasr [INPUT] [OPTIONS]");
    Utils::print ("");
    Utils::print ("Options:");
    Utils::print ("   -h  --help      Display this text.");
    Utils::print ("   -v  --verbose   Verbose mode.");
    Utils::print ("   -o  --output    Output filename (.off format).");
    Utils::print ("   -p  --probes    Number of probes for refining.");
    Utils::print ("   -e  --error     Maximum error for refining.");
    Utils::print ("   -n  --nodes     Number of graph nodes.");
    Utils::print ("   -E  --edges     Number of graph edges.");
    Utils::print ("   -c  --conf      Minimum confidence for sign guess.");
    Utils::print ("   -k  --kmax      Upper bound on local scale for random walk.");
    Utils::print ("   -a  --alpha     Weight for constraints on random walk.");
    Utils::print ("   -A  --angle     Mesh parameter: angle bound.");
    Utils::print ("   -r  --radius    Mesh parameter: radius bound.");
    Utils::print ("   -d  --distance  Mesh parameter: distance bound.");
    Utils::print ("");
    Utils::print ("Default values:");
    Utils::print ("   output    out.off");
    Utils::print ("   probes    10");
    Utils::print ("   error     0.004");
    Utils::print ("   nodes     50000");
    Utils::print ("   edges     1500000");
    Utils::print ("   conf      0.75");
    Utils::print ("   kmax      500");
    Utils::print ("   alpha     1e-3");
    Utils::print ("   angle     30.");
    Utils::print ("   radius    0.05");
    Utils::print ("   distance  0.05");

  }

  void check_validity () const
  {
    if (m_probes < 1)
      Utils::error ("Number of probes should be at least 1.");

    if (m_error <= 0.)
      Utils::error ("Maximum error should be strictly positive.");

    if (m_nodes < 2)
      Utils::error ("Number of nodes should at least 2.");

    if (m_edges < 1)
      Utils::error ("Number of edges should be greater than number of nodes.");

    if (m_conf >= 1.)
      Utils::error ("Minimum confidence should be strictly lower than 1.");

    if (m_kmax < 6)
      Utils::error ("Upper bound on local scale should be at least 6.");
  }

  void display_parameters () const
  {
    Utils::print ("NASR (Noise-Adaptive Shape Reconstruction)");
    Utils::print ("   Verbose mode");
    Utils::print ("");
    Utils::print ("Running with the following arguments:");
    std::cerr << "   * " << m_probes << " probes;" << std::endl;
    std::cerr << "   * Maximum error: " << m_error << std::endl;
    std::cerr << "   * " << m_nodes << " nodes;" << std::endl;
    std::cerr << "   * " << m_edges * m_nodes << " edges;" << std::endl;
    std::cerr << "   * Minimum confidence: " << m_conf << std::endl;
    std::cerr << "   * Upper bound on local scale: " << m_kmax << std::endl;
    std::cerr << "   * Weight for constraints: " << m_alpha << std::endl;
    std::cerr << "   * Mesh angle bound: " << m_mesh_angle << std::endl;
    std::cerr << "   * Mesh radius bound: " << m_mesh_radius << std::endl;
    std::cerr << "   * Mesh distance bound: " << m_mesh_distance << std::endl;
    Utils::print ("");  
    std::cerr << "Input: " << m_input_filename << std::endl;
    std::cerr << "Output: " << m_output_filename << std::endl;
    Utils::print ("");  
  }


  bool verbose () const { return m_verbose; }
  const std::string& input_filename () const { return m_input_filename; }
  const std::string& output_filename () const { return m_output_filename; }
  unsigned int probes () const { return m_probes; }
  double error () const { return m_error; }
  unsigned int nodes () const { return m_nodes; }
  unsigned int edges () const { return m_edges; }
  double conf () const { return m_conf; }
  unsigned int kmax () const { return m_kmax; }
  double alpha () const { return m_alpha; }
  double mesh_angle () const { return m_mesh_angle; }
  double mesh_radius () const { return m_mesh_radius; }
  double mesh_distance () const { return m_mesh_distance; }
};



#endif // PARAMETERS__
