/**
 * NASR - Noise-Adaptive Shape Reconstruction
 * Source code, Version 1.0 (December 2014)
 *
 * Copyright (c) 2014 Simon Giraudot <simon.giraudot@inria.fr>
 * http://www-sop.inria.fr/members/Simon.Giraudot/
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <cstdlib>

#include "common.h"
#include "io_wrapper.h"
#include "parameters.h"


int main (int argc, char** argv)
{
  Parameters parameters (argc, argv);
  IO_wrapper io_wrapper;
  Point_set point_set;
  Iso_cuboid bounding_box;
  Neighbor_search neighbor_search;
  Dt3 dt3;
  Graph graph (dt3);
  Mesh mesh (dt3);



  /////////////////////////////////////////////////////////////////////////////
  // 0. Point set processing
  /////////////////////////////////////////////////////////////////////////////



  Utils::print (parameters.verbose (), "Reading input point set...");
  {
    io_wrapper.read_point_set (parameters.input_filename (), point_set);

    if (parameters.verbose ())
      std::cerr << "Read " << point_set.size () << " points." << std::endl;
  }

  Utils::print (parameters.verbose (), "Normalising point set...");
  {
    Point centroid = CGAL::centroid (point_set.begin (),
				     point_set.end ());

    double max_dist = 0.;
    for (Point_set::iterator it = point_set.begin ();
	 it != point_set.end (); ++ it)
      {
	double dist = CGAL::squared_distance (centroid, *it);
	if (dist > max_dist)
	  max_dist = dist;
      }

    double ratio = 1. / std::sqrt (max_dist);

    for (Point_set::iterator it = point_set.begin ();
	 it != point_set.end (); ++ it)
      {
	Vector vec (centroid, *it);
	*it = Point (0., 0., 0.) +  ratio * vec;
      }
  }

  Utils::print (parameters.verbose (), "Computing bounding box...");
  {
    bounding_box
      = ::CGAL::bounding_box (point_set.begin (), point_set.end ());
  }


  Utils::print (parameters.verbose (), "Computing multiscale kd-tree...");
  {
    neighbor_search.init (point_set.begin (), point_set.end (),
			  point_set.size ());
  }



  /////////////////////////////////////////////////////////////////////////////
  // 1. Noise-adaptive distance function
  /////////////////////////////////////////////////////////////////////////////



  Utils::print (parameters.verbose (), "Refining Delaunay triangulation...");
  {
    Polyhedron bounding_volume;
    CGAL::convex_hull_3 (point_set.begin (), point_set.end (),
			 bounding_volume);

    double ratio = 20000. / (double)(point_set.size ());
    if (ratio > 0.25)
      ratio = 0.25;

    dt3.add_samples_to_triangulation (point_set.begin (),
				      point_set.end (), neighbor_search,
				      ratio, parameters.verbose ());

    dt3.probe_refinement (bounding_volume, 2, parameters.probes (),
			  parameters.error (), neighbor_search, 
			  parameters.verbose ());
  }



  /////////////////////////////////////////////////////////////////////////////
  // 2. Sign guess
  /////////////////////////////////////////////////////////////////////////////



  Utils::print (parameters.verbose (), "Building uniform graph...");
  {
    graph.build_uniform (bounding_box, parameters.nodes ());
  }

  Utils::print (parameters.verbose (), "Picking random edges on graph...");
  {
    graph.pick_random_edges (parameters.edges (), parameters.kmax (),
			     parameters.verbose ());
  }

  Utils::print (parameters.verbose (), "Solving linear system on graph...");
  {
    graph.solve_sign_guess (parameters.verbose ());

    if (parameters.verbose ()) graph.info ();
  }



  /////////////////////////////////////////////////////////////////////////////
  // 3. Reconstruction
  /////////////////////////////////////////////////////////////////////////////



  Utils::print (parameters.verbose (), "Constraining random walker...");
  {
    for (Dt3::Finite_vertices_iterator it = dt3.finite_vertices_begin ();
	 it != dt3.finite_vertices_end (); ++ it)
      it->constrained () = false;

    for (int i = 0; i < graph.size (); ++ i)
      {
	Point p = graph[i].point ();
	Dt3::Vertex_handle v = dt3.nearest_vertex (p);

	if (graph[i].conf () >= parameters.conf ())
	  {
	    v->constrained () = true;
	    v->constraint () = graph[i].sign ();
	  }
      }
  }

  Utils::print (parameters.verbose (), "Building random walker graph...");
  {
    graph.clear ();
    graph.build_on_dt ();
    graph.solve_random_walker (parameters.alpha (),
			       parameters.kmax (),
			       parameters.verbose ());
  }



  /////////////////////////////////////////////////////////////////////////////
  // 4. Mesh
  /////////////////////////////////////////////////////////////////////////////



  Utils::print (parameters.verbose (), "Generating mesh...");
  {

    dt3.switch_signs_if_opposite (parameters.verbose ());

    double median_value 
      = dt3.median_implicit_at_input_points (point_set.begin (),
      					     point_set.end ());

    if (parameters.verbose ())
      std::cout << "   Median value of implicit function on input points = "
		<< median_value << std::endl;

    Point center = dt3.get_random_point_with_implicit_negative
      (median_value);

    double size = 0;
    for (int i = 0; i < point_set.size (); ++ i)
      {
	double dist = CGAL::squared_distance (center, point_set[i]);
	if (dist > size)
	  size = dist;
      }

    size *= 2;

    Sphere s (center, size);

    mesh.generate (median_value, s,
		   parameters.mesh_angle (),
		   parameters.mesh_radius (),
		   parameters.mesh_distance ());
  }

  Utils::print (parameters.verbose (), "Writing output surface...");
  {
    io_wrapper.write_surface (parameters.output_filename (), mesh);
  }

  Utils::print (parameters.verbose (), "All done. Exiting.");

  return EXIT_SUCCESS;
}






